//
//  ASObject.m
//  Paste2
//
//  Created by Andrew Finnell on 10/28/09.
//  Copyright 2009 Order N Development, LLC. All rights reserved.
//

#import "ASObject.h"
#import "ASElementArray.h"

NSString * const ASObjectAppleScriptException = @"ASObjectAppleScriptException";

NSString *NSStringFromAppleEvent(AppleEvent event)
{
	Handle printableHandle = nil;
	AEPrintDescToHandle(&event, &printableHandle);
	NSString *string = [[[NSString alloc] initWithBytes:*printableHandle length:GetHandleSize(printableHandle) encoding:NSUTF8StringEncoding] autorelease];
	DisposeHandle(printableHandle);
	return string;
}

OSType NSStringToOSType(NSString *key)
{
	assert( [key length] == 4 );
	
	OSType type = 0;
	char *chars = (char *)&type;
	
	for (int i = 0; i < 4; ++i)
		chars[i] = [key characterAtIndex:i];
	
	return CFSwapInt32BigToHost(type);
}

static NSAppleEventDescriptor *ObjectToAEDesc(id object, NSString *asType)
{
	NSAppleEventDescriptor *desc = nil;
	
	if ( [object isKindOfClass:[NSNumber class]] ) {		
		if ( [asType isEqualToString:@"type"] )
			desc = [NSAppleEventDescriptor descriptorWithTypeCode:[object unsignedIntValue]];
		else {
			const char *objCType = [object objCType];
			if ( strcmp(objCType, @encode(char)) == 0 )
				desc = [NSAppleEventDescriptor descriptorWithBoolean:[object boolValue]];
			else				
				desc = [NSAppleEventDescriptor descriptorWithInt32:[object intValue]];
		}
	} else if ( [object isKindOfClass:[NSString class]] ) 
		desc = [NSAppleEventDescriptor descriptorWithString:object];
	else if ( [object isKindOfClass:[NSData class]] )
		desc = [NSAppleEventDescriptor descriptorWithDescriptorType:'tdta' data:object];
	else if ( [object isKindOfClass:[ASObject class]] )
		desc = [[object objectSpecifier] descriptor];
	else if ( [object isKindOfClass:[NSAppleEventDescriptor class]] )
		desc = object;
	else if ( object == [NSNull null] )
		desc = [NSAppleEventDescriptor descriptorWithTypeCode:'msng'];
	else if ( [object isKindOfClass:[NSDictionary class]] ) {
		NSDictionary *dictionary = object;
		desc = [NSAppleEventDescriptor recordDescriptor];
		for (NSString *key in dictionary) {
			NSAppleEventDescriptor *valueDesc = ObjectToAEDesc([dictionary objectForKey:key], @"any");
			[desc setDescriptor:valueDesc forKeyword:NSStringToOSType(key)];
		}		
	} else 
		NSLog(@"unknown object of as type: %@", asType);
	
	return desc;
}

static NSAppleEventDescriptor *NSDictionaryToAEDesc(NSDictionary *properties, NSScriptClassDescription *description)
{
	NSAppleEventDescriptor *record = [NSAppleEventDescriptor recordDescriptor];
	
	for (NSString *key in properties) {
		FourCharCode code = [description appleEventCodeForKey:key];
		id value = [properties objectForKey:key];
		NSString *type = [description typeForKey:key];
		
		[record setDescriptor:ObjectToAEDesc(value, type) forKeyword:code];
	}
	
	return record;
}

static ASObject *ASObjFromReply(AppleEvent reply)
{
	NSAppleEventDescriptor *appleEvent = [[[NSAppleEventDescriptor alloc] initWithAEDescNoCopy:&reply] autorelease];
	NSAppleEventDescriptor *directObject = [appleEvent paramDescriptorForKeyword:keyDirectObject];
	
	if ( directObject != nil )
		return [[[ASObject alloc] initWithScriptObjectSpecifier:[NSScriptObjectSpecifier objectSpecifierWithDescriptor:directObject]] autorelease];
				
	// Something went bad. Look for error information
	NSAppleEventDescriptor *errorNumber = [appleEvent paramDescriptorForKeyword:keyErrorNumber];
	NSAppleEventDescriptor *errorString = [appleEvent paramDescriptorForKeyword:keyErrorString];

	@throw [NSException exceptionWithName:ASObjectAppleScriptException reason:[errorString stringValue] userInfo:[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithInt:[errorNumber int32Value]], @"number", [errorString stringValue], @"string", nil]];
}

id ObjectFromReply(AppleEvent reply)
{
	NSAppleEventDescriptor *appleEvent = [[[NSAppleEventDescriptor alloc] initWithAEDescNoCopy:&reply] autorelease];
	NSAppleEventDescriptor *directObject = [appleEvent paramDescriptorForKeyword:keyDirectObject];
	
	if ( directObject != nil ) {
		DescType type = [directObject descriptorType];
		
		switch (type) {
			case typeObjectSpecifier:
				return [[[ASObject alloc] initWithScriptObjectSpecifier:[NSScriptObjectSpecifier objectSpecifierWithDescriptor:directObject]] autorelease];
				break;
			
			case typeUTF8Text:
			case typeUnicodeText:
			case typeChar:
				return [directObject stringValue];
				break;
			
			case typeSInt16:
			case typeSInt32:
				return [NSNumber numberWithInt:[directObject int32Value]];
				break;
				
			case typeBoolean:
				return [NSNumber numberWithBool:[directObject booleanValue]];
				break;
			
			case typeData:
				return [directObject data];
				break;
				
			case 'msng':
				return nil;
				break;
				
			default:
				NSLog(@"unsupported DescType %u", (unsigned)type);
				break;
		}
	}
	
	// Look for error information to see if anything went wrong. Could just be a void method
	NSAppleEventDescriptor *errorNumber = [appleEvent paramDescriptorForKeyword:keyErrorNumber];
	NSAppleEventDescriptor *errorString = [appleEvent paramDescriptorForKeyword:keyErrorString];
	
	if ( errorNumber != nil || errorString != nil )
		@throw [NSException exceptionWithName:ASObjectAppleScriptException reason:[errorString stringValue] userInfo:[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithInt:[errorNumber int32Value]], @"number", [errorString stringValue], @"string", nil]];	

	return nil;
}


@implementation ASObject

+ (id) makeNewClass:(Class)theClass withProperties:(NSDictionary *)properties withData:(id)data inContainer:(ASElementArray *)container atIndex:(int)index
{
	NSScriptClassDescription *description = [NSScriptClassDescription classDescriptionForClass:theClass];
	NSAppleEventDescriptor *specifier = [container insertionLocationForIndex:index];
	NSAppleEventDescriptor *propertiesDesc = NSDictionaryToAEDesc(properties, description);

	return [ASObject invokeCommand:@"make" withParameters:[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedInt:[description appleEventCode]], @"ObjectClass", 
														   specifier, @"Location",
														   propertiesDesc, @"KeyDictionary",
														   data, @"ObjectData",
														   nil]];

}

- (id) initWithScriptObjectSpecifier:(NSScriptObjectSpecifier *)specifier
{
	self = [super init];
	
	if ( self != nil ) {
		_specifier = [specifier retain];
	}

	return self;
}

- (void) dealloc
{
	[_specifier release];
	
	[super dealloc];
}

- (NSScriptClassDescription *) scriptClassDescription
{
	return [_specifier keyClassDescription];
}

- (NSScriptObjectSpecifier *) objectSpecifier
{
	return [[_specifier retain] autorelease];
}

- (ASElementArray *) elementForKey:(NSString *)string
{
	return [[[ASElementArray alloc] initWithContainer:self key:string] autorelease];
}

- (void) setObject:(id)value forProperty:(NSString *)property
{
	NSScriptObjectSpecifier *specifier = [[[NSPropertySpecifier alloc] initWithContainerSpecifier:_specifier key:property] autorelease];
	[ASObject invokeCommand:@"set" withParameters:[NSDictionary dictionaryWithObjectsAndKeys:[specifier descriptor], @"", value, @"Value", nil]];
}

- (id) objectForProperty:(NSString *)property
{
	NSScriptObjectSpecifier *specifier = [[[NSPropertySpecifier alloc] initWithContainerSpecifier:_specifier key:property] autorelease];
	return [ASObject invokeCommand:@"get" withParameters:[NSDictionary dictionaryWithObjectsAndKeys:[specifier descriptor], @"", nil]];
}

- (void) delete
{
	[self invokeCommand:@"delete"];
}

+ (id) invokeCommand:(NSString *)name with:(id)firstObject, ...
{
	va_list list = {};
	
	va_start(list, firstObject);
	
	NSMutableDictionary *parameters = [[[NSMutableDictionary alloc] init] autorelease];
	id object = firstObject;
	while ( object != nil ) {
		NSString *key = va_arg(list, NSString *);
		[parameters setObject:object forKey:key];
		
		object = va_arg(list, id);
	}
	
	va_end(list);
	
	return [ASObject invokeCommand:name withParameters:parameters];	
}

+ (id) invokeCommand:(NSString *)name
{
	return [ASObject invokeCommand:name withParameters:[NSDictionary dictionary]];
}

+ (id) invokeCommand:(NSString *)name withParameters:(NSDictionary *)parameters
{
	return [ASObject invokeCommand:name inSuite:nil onObject:nil withParameters:parameters];
}

+ (id) invokeCommand:(NSString *)name inSuite:(NSString *)suiteName onObject:(ASObject *)object withParameters:(NSDictionary *)parameters
{
	// First, find the command description
	NSScriptCommandDescription *commandDescription = nil;
	NSScriptSuiteRegistry *registry = [NSScriptSuiteRegistry sharedScriptSuiteRegistry];
	
	if ( suiteName != nil ) {
		NSDictionary *commands = [registry commandDescriptionsInSuite:suiteName];
		commandDescription = [commands objectForKey:name];
	} 
	if ( commandDescription == nil ) {
		NSArray *suites = [registry suiteNames];
		
		for (NSString *theSuite in suites) {
			NSDictionary *commands = [registry commandDescriptionsInSuite:theSuite];
			commandDescription = [commands objectForKey:name];
			if ( commandDescription != nil )
				break;
		}
	}
	
	if ( commandDescription == nil )
		@throw [NSException exceptionWithName:ASObjectAppleScriptException reason:[NSString stringWithFormat:@"Could not find command %@", name] userInfo:nil];
	
	// Create an apple event with the proper target
	ProcessSerialNumber psn = { 0, kCurrentProcess };
	NSAppleEventDescriptor *target = [NSAppleEventDescriptor descriptorWithDescriptorType:typeProcessSerialNumber bytes:&psn length:sizeof(psn)];
	NSAppleEventDescriptor *event = [NSAppleEventDescriptor appleEventWithEventClass:[commandDescription appleEventClassCode] eventID:[commandDescription appleEventCode] targetDescriptor:target returnID:kAutoGenerateReturnID transactionID:kAnyTransactionID];

	// Add parameters
	if ( object != nil )
		[event setParamDescriptor:[[object objectSpecifier] descriptor] forKeyword:keyDirectObject];
	for (NSString *parameterName in parameters) {
		id parameterValue = [parameters objectForKey:parameterName];

		// Special case for direct object
		if ( [parameterName length] == 0 ) {
			[event setParamDescriptor:parameterValue forKeyword:keyDirectObject];
			continue;
		}
		
		NSString *parameterType = [commandDescription typeForArgumentWithName:parameterName];
		
		NSAppleEventDescriptor *parameterDesc = ObjectToAEDesc(parameterValue, parameterType);
		FourCharCode parameterCode = [commandDescription appleEventCodeForArgumentWithName:parameterName];
		[event setParamDescriptor:parameterDesc forKeyword:parameterCode];
	}
	
	// Actually invoke the command
	AppleEvent reply = {};
	AESendMessage([event aeDesc], &reply, kAEWaitReply, kAEDefaultTimeout);
	
	return ObjectFromReply(reply);	
}

- (id) invokeCommand:(NSString *)name withParameters:(NSDictionary *)parameters
{
	return [ASObject invokeCommand:name inSuite:[[self scriptClassDescription] suiteName] onObject:self withParameters:parameters];
}

- (id) invokeCommand:(NSString *)name with:(id)firstObject, ...
{
	va_list list = {};
	
	va_start(list, firstObject);
	
	NSMutableDictionary *parameters = [[[NSMutableDictionary alloc] init] autorelease];
	id object = firstObject;
	while ( object != nil ) {
		NSString *key = va_arg(list, NSString *);
		[parameters setObject:object forKey:key];
		
		object = va_arg(list, id);
	}
	
	va_end(list);
	
	return [self invokeCommand:name withParameters:parameters];
}

- (id) invokeCommand:(NSString *)name
{
	return [self invokeCommand:name withParameters:[NSDictionary dictionary]];
}

@end

ASObject *ASObj(id obj)
{
	NSScriptObjectSpecifier *specifier = [obj objectSpecifier];
	if ( specifier == nil )
		return nil;
	return [[[ASObject alloc] initWithScriptObjectSpecifier:specifier] autorelease];
}
